# LICENSE: PSF.

import asyncio
import concurrent.futures
import re
import sys
import threading
import unittest
import uvloop

from asyncio import test_utils
from uvloop import _testbase as tb
from unittest import mock
from test import support


# Most of the tests are copied from asyncio


def _fakefunc(f):
    return f


def first_cb():
    pass


def last_cb():
    pass


class _TestFutures:

    def create_future(self):
        raise NotImplementedError

    def test_future_initial_state(self):
        f = self.create_future()
        self.assertFalse(f.cancelled())
        self.assertFalse(f.done())
        f.cancel()
        self.assertTrue(f.cancelled())

    def test_future_cancel(self):
        f = self.create_future()
        self.assertTrue(f.cancel())
        self.assertTrue(f.cancelled())
        self.assertTrue(f.done())
        self.assertRaises(asyncio.CancelledError, f.result)
        self.assertRaises(asyncio.CancelledError, f.exception)
        self.assertRaises(asyncio.InvalidStateError, f.set_result, None)
        self.assertRaises(asyncio.InvalidStateError, f.set_exception, None)
        self.assertFalse(f.cancel())

    def test_future_result(self):
        f = self.create_future()
        self.assertRaises(asyncio.InvalidStateError, f.result)

        f.set_result(42)
        self.assertFalse(f.cancelled())
        self.assertTrue(f.done())
        self.assertEqual(f.result(), 42)
        self.assertEqual(f.exception(), None)
        self.assertRaises(asyncio.InvalidStateError, f.set_result, None)
        self.assertRaises(asyncio.InvalidStateError, f.set_exception, None)
        self.assertFalse(f.cancel())

    def test_future_exception(self):
        exc = RuntimeError()
        f = self.create_future()
        self.assertRaises(asyncio.InvalidStateError, f.exception)

        if sys.version_info[:3] > (3, 5, 1):
            # StopIteration cannot be raised into a Future - CPython issue26221
            self.assertRaisesRegex(TypeError,
                                   "StopIteration .* cannot be raised",
                                   f.set_exception, StopIteration)

        f.set_exception(exc)
        self.assertFalse(f.cancelled())
        self.assertTrue(f.done())
        self.assertRaises(RuntimeError, f.result)
        self.assertEqual(f.exception(), exc)
        self.assertRaises(asyncio.InvalidStateError, f.set_result, None)
        self.assertRaises(asyncio.InvalidStateError, f.set_exception, None)
        self.assertFalse(f.cancel())

    def test_future_exception_class(self):
        f = self.create_future()
        f.set_exception(RuntimeError)
        self.assertIsInstance(f.exception(), RuntimeError)

    def test_future_yield_from_twice(self):
        f = self.create_future()

        def fixture():
            yield 'A'
            x = yield from f
            yield 'B', x
            y = yield from f
            yield 'C', y

        g = fixture()
        self.assertEqual(next(g), 'A')  # yield 'A'.
        self.assertEqual(next(g), f)  # First yield from f.
        f.set_result(42)
        self.assertEqual(next(g), ('B', 42))  # yield 'B', x.
        # The second "yield from f" does not yield f.
        self.assertEqual(next(g), ('C', 42))  # yield 'C', y.

    def test_future_repr(self):
        self.loop.set_debug(True)
        f_pending_debug = self.create_future()
        frame = f_pending_debug._source_traceback[-1]
        self.assertEqual(repr(f_pending_debug),
                         '<Future pending created at %s:%s>'
                         % (frame[0], frame[1]))
        f_pending_debug.cancel()

        self.loop.set_debug(False)
        f_pending = self.create_future()
        self.assertEqual(repr(f_pending), '<Future pending>')
        f_pending.cancel()

        f_cancelled = self.create_future()
        f_cancelled.cancel()
        self.assertEqual(repr(f_cancelled), '<Future cancelled>')

        f_result = self.create_future()
        f_result.set_result(4)
        self.assertEqual(repr(f_result), '<Future finished result=4>')
        self.assertEqual(f_result.result(), 4)

        exc = RuntimeError()
        f_exception = self.create_future()
        f_exception.set_exception(exc)
        self.assertEqual(repr(f_exception),
                         '<Future finished exception=RuntimeError()>')
        self.assertIs(f_exception.exception(), exc)

        def func_repr(func):
            filename, lineno = test_utils.get_function_source(func)
            text = '%s() at %s:%s' % (func.__qualname__, filename, lineno)
            return re.escape(text)

        f_one_callbacks = self.create_future()
        f_one_callbacks.add_done_callback(_fakefunc)
        fake_repr = func_repr(_fakefunc)
        self.assertRegex(repr(f_one_callbacks),
                         r'<Future pending cb=\[%s\]>' % fake_repr)
        f_one_callbacks.cancel()
        self.assertEqual(repr(f_one_callbacks),
                         '<Future cancelled>')

        f_two_callbacks = self.create_future()
        f_two_callbacks.add_done_callback(first_cb)
        f_two_callbacks.add_done_callback(last_cb)
        first_repr = func_repr(first_cb)
        last_repr = func_repr(last_cb)
        self.assertRegex(repr(f_two_callbacks),
                         r'<Future pending cb=\[%s, %s\]>'
                         % (first_repr, last_repr))

        f_many_callbacks = self.create_future()
        f_many_callbacks.add_done_callback(first_cb)
        for i in range(8):
            f_many_callbacks.add_done_callback(_fakefunc)
        f_many_callbacks.add_done_callback(last_cb)
        cb_regex = r'%s, <8 more>, %s' % (first_repr, last_repr)
        self.assertRegex(repr(f_many_callbacks),
                         r'<Future pending cb=\[%s\]>' % cb_regex)
        f_many_callbacks.cancel()
        self.assertEqual(repr(f_many_callbacks),
                         '<Future cancelled>')

    def test_future_copy_state(self):
        if sys.version_info[:3] < (3, 5, 1):
            raise unittest.SkipTest()

        from asyncio.futures import _copy_future_state

        f = self.create_future()
        f.set_result(10)

        newf = self.create_future()
        _copy_future_state(f, newf)
        self.assertTrue(newf.done())
        self.assertEqual(newf.result(), 10)

        f_exception = self.create_future()
        f_exception.set_exception(RuntimeError())

        newf_exception = self.create_future()
        _copy_future_state(f_exception, newf_exception)
        self.assertTrue(newf_exception.done())
        self.assertRaises(RuntimeError, newf_exception.result)

        f_cancelled = self.create_future()
        f_cancelled.cancel()

        newf_cancelled = self.create_future()
        _copy_future_state(f_cancelled, newf_cancelled)
        self.assertTrue(newf_cancelled.cancelled())

    @mock.patch('asyncio.base_events.logger')
    def test_future_tb_logger_abandoned(self, m_log):
        fut = self.create_future()
        del fut
        self.assertFalse(m_log.error.called)

    @mock.patch('asyncio.base_events.logger')
    def test_future_tb_logger_result_unretrieved(self, m_log):
        fut = self.create_future()
        fut.set_result(42)
        del fut
        self.assertFalse(m_log.error.called)

    @mock.patch('asyncio.base_events.logger')
    def test_future_tb_logger_result_retrieved(self, m_log):
        fut = self.create_future()
        fut.set_result(42)
        fut.result()
        del fut
        self.assertFalse(m_log.error.called)

    def test_future_wrap_future(self):
        def run(arg):
            return (arg, threading.get_ident())
        ex = concurrent.futures.ThreadPoolExecutor(1)
        f1 = ex.submit(run, 'oi')
        f2 = asyncio.wrap_future(f1, loop=self.loop)
        res, ident = self.loop.run_until_complete(f2)
        self.assertIsInstance(f2, asyncio.Future)
        self.assertEqual(res, 'oi')
        self.assertNotEqual(ident, threading.get_ident())

    def test_future_wrap_future_future(self):
        f1 = self.create_future()
        f2 = asyncio.wrap_future(f1)
        self.assertIs(f1, f2)

    def test_future_wrap_future_use_global_loop(self):
        with mock.patch('asyncio.futures.events') as events:
            events.get_event_loop = lambda: self.loop
            def run(arg):
                return (arg, threading.get_ident())
            ex = concurrent.futures.ThreadPoolExecutor(1)
            f1 = ex.submit(run, 'oi')
            f2 = asyncio.wrap_future(f1)
            self.assertIs(self.loop, f2._loop)

    def test_future_wrap_future_cancel(self):
        f1 = concurrent.futures.Future()
        f2 = asyncio.wrap_future(f1, loop=self.loop)
        f2.cancel()
        test_utils.run_briefly(self.loop)
        self.assertTrue(f1.cancelled())
        self.assertTrue(f2.cancelled())

    def test_future_wrap_future_cancel2(self):
        f1 = concurrent.futures.Future()
        f2 = asyncio.wrap_future(f1, loop=self.loop)
        f1.set_result(42)
        f2.cancel()
        test_utils.run_briefly(self.loop)
        self.assertFalse(f1.cancelled())
        self.assertEqual(f1.result(), 42)
        self.assertTrue(f2.cancelled())

    def test_future_source_traceback(self):
        self.loop.set_debug(True)

        future = self.create_future()
        lineno = sys._getframe().f_lineno - 1
        self.assertIsInstance(future._source_traceback, list)
        self.assertEqual(future._source_traceback[-2][:3],
                         (__file__,
                          lineno,
                          'test_future_source_traceback'))

    def check_future_exception_never_retrieved(self, debug):
        last_ctx = None
        def handler(loop, context):
            nonlocal last_ctx
            last_ctx = context

        self.loop.set_debug(debug)
        self.loop.set_exception_handler(handler)

        def memory_error():
            try:
                raise MemoryError()
            except BaseException as exc:
                return exc
        exc = memory_error()

        future = self.create_future()
        if debug:
            source_traceback = future._source_traceback
        future.set_exception(exc)
        future = None
        support.gc_collect()
        test_utils.run_briefly(self.loop)

        self.assertIsNotNone(last_ctx)

        self.assertIs(last_ctx['exception'], exc)
        self.assertEqual(last_ctx['message'],
                         'Future exception was never retrieved')

        if debug:
            tb = last_ctx['source_traceback']
            self.assertEqual(tb[-2].name,
                             'check_future_exception_never_retrieved')

    def test_future_exception_never_retrieved(self):
        self.check_future_exception_never_retrieved(False)

    def test_future_exception_never_retrieved_debug(self):
        self.check_future_exception_never_retrieved(True)

    def test_future_wrap_future(self):
        from uvloop.loop import _wrap_future
        def run(arg):
            return (arg, threading.get_ident())
        ex = concurrent.futures.ThreadPoolExecutor(1)
        f1 = ex.submit(run, 'oi')
        f2 = _wrap_future(f1, loop=self.loop)
        res, ident = self.loop.run_until_complete(f2)
        self.assertIsInstance(f2, asyncio.Future)
        self.assertEqual(res, 'oi')
        self.assertNotEqual(ident, threading.get_ident())

    def test_future_wrap_future_future(self):
        from uvloop.loop import _wrap_future
        f1 = self.create_future()
        f2 = _wrap_future(f1)
        self.assertIs(f1, f2)

    def test_future_wrap_future_cancel(self):
        from uvloop.loop import _wrap_future
        f1 = concurrent.futures.Future()
        f2 = _wrap_future(f1, loop=self.loop)
        f2.cancel()
        test_utils.run_briefly(self.loop)
        self.assertTrue(f1.cancelled())
        self.assertTrue(f2.cancelled())

    def test_future_wrap_future_cancel2(self):
        from uvloop.loop import _wrap_future
        f1 = concurrent.futures.Future()
        f2 = _wrap_future(f1, loop=self.loop)
        f1.set_result(42)
        f2.cancel()
        test_utils.run_briefly(self.loop)
        self.assertFalse(f1.cancelled())
        self.assertEqual(f1.result(), 42)
        self.assertTrue(f2.cancelled())


class _TestFuturesDoneCallbacks:

    def run_briefly(self):
        test_utils.run_briefly(self.loop)

    def _make_callback(self, bag, thing):
        # Create a callback function that appends thing to bag.
        def bag_appender(future):
            bag.append(thing)
        return bag_appender

    def _new_future(self):
        raise NotImplementedError

    def test_future_callbacks_invoked_on_set_result(self):
        bag = []
        f = self._new_future()
        f.add_done_callback(self._make_callback(bag, 42))
        f.add_done_callback(self._make_callback(bag, 17))

        self.assertEqual(bag, [])
        f.set_result('foo')

        self.run_briefly()

        self.assertEqual(bag, [42, 17])
        self.assertEqual(f.result(), 'foo')

    def test_future_callbacks_invoked_on_set_exception(self):
        bag = []
        f = self._new_future()
        f.add_done_callback(self._make_callback(bag, 100))

        self.assertEqual(bag, [])
        exc = RuntimeError()
        f.set_exception(exc)

        self.run_briefly()

        self.assertEqual(bag, [100])
        self.assertEqual(f.exception(), exc)

    def test_future_remove_done_callback(self):
        bag = []
        f = self._new_future()
        cb1 = self._make_callback(bag, 1)
        cb2 = self._make_callback(bag, 2)
        cb3 = self._make_callback(bag, 3)

        # Add one cb1 and one cb2.
        f.add_done_callback(cb1)
        f.add_done_callback(cb2)

        # One instance of cb2 removed. Now there's only one cb1.
        self.assertEqual(f.remove_done_callback(cb2), 1)

        # Never had any cb3 in there.
        self.assertEqual(f.remove_done_callback(cb3), 0)

        # After this there will be 6 instances of cb1 and one of cb2.
        f.add_done_callback(cb2)
        for i in range(5):
            f.add_done_callback(cb1)

        # Remove all instances of cb1. One cb2 remains.
        self.assertEqual(f.remove_done_callback(cb1), 6)

        self.assertEqual(bag, [])
        f.set_result('foo')

        self.run_briefly()

        self.assertEqual(bag, [2])
        self.assertEqual(f.result(), 'foo')


###############################################################################
# Tests Matrix
###############################################################################


class Test_UV_UV_create_future(_TestFutures, tb.UVTestCase):
    # Test uvloop.Loop.create_future
    def create_future(self):
        return self.loop.create_future()


class Test_UV_UV_Future(_TestFutures, tb.UVTestCase):
    # Test that uvloop.Future can be instantiated directly
    def create_future(self):
        return uvloop.Future(loop=self.loop)


class Test_UV_AIO_Futures(_TestFutures, tb.UVTestCase):
    def create_future(self):
        return asyncio.Future(loop=self.loop)


class Test_AIO_Futures(_TestFutures, tb.AIOTestCase):
    def create_future(self):
        return asyncio.Future(loop=self.loop)


class Test_UV_UV_FuturesCallbacks(_TestFuturesDoneCallbacks, tb.UVTestCase):
    def _new_future(self):
        return self.loop.create_future()


class Test_UV_AIO_FuturesCallbacks(_TestFuturesDoneCallbacks, tb.UVTestCase):
    def _new_future(self):
        return asyncio.Future(loop=self.loop)


class Test_AIO_FuturesCallbacks(_TestFuturesDoneCallbacks, tb.AIOTestCase):
    def _new_future(self):
        return asyncio.Future(loop=self.loop)



# LICENSE: PSF.

DEF _FUT_PENDING = 1
DEF _FUT_CANCELLED = 2
DEF _FUT_FINISHED = 3


cdef inline _future_get_blocking(fut):
    try:
        return fut._asyncio_future_blocking
    except AttributeError:
        return fut._blocking


cdef inline _future_set_blocking(fut, val):
    try:
        fut._asyncio_future_blocking
    except AttributeError:
        fut._blocking = val
    else:
        fut._asyncio_future_blocking = val


cdef class BaseFuture:
    cdef:
        int _state

        readonly Loop _loop
        readonly list _callbacks
        readonly object _exception
        readonly object _result
        public bint _blocking
        readonly object _source_traceback
        readonly bint _log_traceback

    def __init__(self, Loop loop):
        if loop is None:
            loop = aio_get_event_loop()
            if not isinstance(loop, Loop):
                raise TypeError('uvloop.Future supports only uvloop.Loop')

        self._state = _FUT_PENDING
        self._loop = loop
        self._callbacks = []
        self._result = None
        self._exception = None
        self._blocking = False
        self._log_traceback = False

        if loop._debug:
            self._source_traceback = tb_extract_stack(sys_getframe(0))
        else:
            self._source_traceback = None

    property _asyncio_future_blocking:
        def __get__(self):
            return self._blocking

        def __set__(self, value):
            self._blocking = value

    cdef _schedule_callbacks(self):
        cdef:
            list callbacks
            size_t cb_len = len(self._callbacks)
            size_t i

        if cb_len == 0:
            return

        callbacks = self._callbacks
        self._callbacks = []

        for i from 0 <= i < cb_len:
            self._loop.call_soon(callbacks[i], self)

    cdef _add_done_callback(self, fn):
        if self._state != _FUT_PENDING:
            self._loop.call_soon(fn, self)
        else:
            self._callbacks.append(fn)

    cdef _done(self):
        return self._state != _FUT_PENDING

    cdef _cancel(self):
        if self._done():
            return False
        self._state = _FUT_CANCELLED
        self._schedule_callbacks()
        return True

    # _result would shadow the "_result" property
    cdef _result_impl(self):
        if self._state == _FUT_CANCELLED:
            raise aio_CancelledError
        if self._state != _FUT_FINISHED:
            raise aio_InvalidStateError('Result is not ready.')
        self._log_traceback = False
        if self._exception is not None:
            raise self._exception
        return self._result

    cdef _str_state(self):
        if self._state == _FUT_PENDING:
            return 'PENDING'
        elif self._state == _FUT_CANCELLED:
            return 'CANCELLED'
        elif self._state == _FUT_FINISHED:
            return 'FINISHED'
        else:
            raise RuntimeError('unknown Future state')

    property _state:
        def __get__(self):
            return self._str_state()

    def cancel(self):
        """Cancel the future and schedule callbacks.

        If the future is already done or cancelled, return False.  Otherwise,
        change the future's state to cancelled, schedule the callbacks and
        return True.
        """
        return self._cancel()

    def cancelled(self):
        """Return True if the future was cancelled."""
        return self._state == _FUT_CANCELLED

    def done(self):
        """Return True if the future is done.

        Done means either that a result / exception are available, or that the
        future was cancelled.
        """
        return self._state != _FUT_PENDING

    def result(self):
        """Return the result this future represents.

        If the future has been cancelled, raises CancelledError.  If the
        future's result isn't yet available, raises InvalidStateError.  If
        the future is done and has an exception set, this exception is raised.
        """
        return self._result_impl()

    def exception(self):
        """Return the exception that was set on this future.

        The exception (or None if no exception was set) is returned only if
        the future is done.  If the future has been cancelled, raises
        CancelledError.  If the future isn't done yet, raises
        InvalidStateError.
        """
        if self._state == _FUT_CANCELLED:
            raise aio_CancelledError
        if self._state != _FUT_FINISHED:
            raise aio_InvalidStateError('Exception is not set.')
        self._log_traceback = False
        return self._exception

    def add_done_callback(self, fn):
        """Add a callback to be run when the future becomes done.

        The callback is called with a single argument - the future object. If
        the future is already done when this is called, the callback is
        scheduled with call_soon.
        """
        self._add_done_callback(fn)

    def remove_done_callback(self, fn):
        """Remove all instances of a callback from the "call when done" list.

        Returns the number of callbacks removed.
        """
        cdef:
            size_t clen = len(self._callbacks)
            size_t i
            size_t ni = 0
            object cb

        for i from 0 <= i < clen:
            cb = self._callbacks[i]
            if cb != fn:
                self._callbacks[ni] = cb
                ni += 1

        if ni != clen:
            del self._callbacks[ni:]

        return clen - ni

    cpdef set_result(self, result):
        """Mark the future done and set its result.

        If the future is already done when this method is called, raises
        InvalidStateError.
        """
        if self._state != _FUT_PENDING:
            raise aio_InvalidStateError('{}: {!r}'.format(
                self._str_state(), self))
        self._result = result
        self._state = _FUT_FINISHED
        self._schedule_callbacks()

    cpdef set_exception(self, exception):
        """Mark the future done and set an exception.

        If the future is already done when this method is called, raises
        InvalidStateError.
        """
        if self._state != _FUT_PENDING:
            raise aio_InvalidStateError('{}: {!r}'.format(
                self._str_state(), self))
        if isinstance(exception, type):
            exception = exception()
        if type(exception) is StopIteration:
            raise TypeError("StopIteration interacts badly with generators "
                            "and cannot be raised into a Future")
        self._exception = exception
        self._state = _FUT_FINISHED
        self._schedule_callbacks()
        self._log_traceback = True

    # Copy of __await__
    def __iter__(self):
        if self._state == _FUT_PENDING:
            self._blocking = True
            yield self  # This tells Task to wait for completion.

        if self._state == _FUT_PENDING:
            raise AssertionError("yield from wasn't used with future")

        return self._result_impl()  # May raise too.

    # Copy of __iter__
    def __await__(self):
        if self._state == _FUT_PENDING:
            self._blocking = True
            yield self  # This tells Task to wait for completion.

        if self._state == _FUT_PENDING:
            raise AssertionError("yield from wasn't used with future")

        return self._result_impl()  # May raise too.


class Future(BaseFuture, aio_Future):
    # Inherit asyncio.Future.__del__ and __repr__
    pass


cdef uvloop_Future = Future
